%:fil QTCOMF.SL 	FUNCTION/Object LIBRARY for TCOMP	%:hlp
%: You may in your input use <DEL A B>, <RDEL A B>, <EPS2 A B>,<EPS3 A B C>,
%: <EPS4 A B C D>,...<EPS9 A B ... G H I>, se manual.

% There might also be other lisp code put in the beginning of a .BLF-file, that
% defines functions needed for computing some of the tensors TCOMPed hitherto.
% These functions are not always needed, as the functions in SUMBOU are.
% Therefore they are instead put in the beginning of the .BLF file, but only if
% they have been found to be necessary for any of the tensors TCOMPed hitherto.

%	If you TCOMPOUT two files of different tensors, then the 2:nd might
%	contain some routines only needed in the first. This slight waste
%	is most easily avoided by restarting STENSOR for the 2:nd file.

%	It is TCOMP that with PRQFN adds them to the accumulated list PREBLF!*,
%	as they are found to be needed. In case TCOMP is used before, then
%	TCOMPOUT won't regenerate the bulf, but save the one already produced.
%	Therefore TCOMPOUT doesn't reset PREBLF!*, but has to continue to add
%	to the list the previous TCOMP started on, ie PREBLF!* is global.

%	Unfortunately, if you TCOMPOUT two files, the 2:nd will then also
%	save routines only needed in the 1:st. Avoided by not using TCOMP,
%	and doing (SETQ PREBLF!* '(NIL)) in between TCOMPOUTs.
%	But if file 2 shares a sub-hierarchy of tensors with file 1, those
%	are already TCOMPed at 2:nd TCOMPOUT. Then keep old PREBLF!*, or better
%	(re)load file2.DEF (the PDEFs rem's the BULFs with XBULF).

% COMPILATION:
%	In larger packages, you might wish to move (some of) them to a separate
%	(compiled) file. Then load that also to STENSOR, and TCOMPOUT won't put
%	them to .BLF, only warn "Preq-fn compiled,can't save it. Name-clash?".

% **** FUNCTION LIBRARY that was discussed above follows:  ****

(COND (T	% Avoid compilation! (So can be GRINDed into .BLF-file)
		% vvvv START OF COND vvvv
  (DE EDTM!? (X) 
   (COND 
     ((AND % (PAIRP X)			% EQCAR supposed to do check!
	   (EQCAR X 'TIMES)		% VAXDEC??
	   (EQ (CADR X) -1)
	   (NOT (CDDDR X)))
      (FUNCALL 'SMMINUS (CADDR X)) )		%  'EDITM on DEC
     (X) ))

%+Slisp
  (DM MKTMN (L) (EXPAND (CDR L) 'MKTIMS))
  (VAXDEC NIL (DM XMPLUS (L) (EXPAND (CDR L) 'PLUSIM)))
%-Slisp

%if SuM then SyMBOL: if X sum then return symbol A. Used in neg powers eg 1/det
% (DE SMBOL (A X) (OR (AND (EQCAR X 'PLUS) A) (COPYL X)))	%:PSL gave NIL
% (DE SMBOL (A X) (COND ((EQCAR X 'PLUS) A) (T (COPYL X))))	% Dont compile
  (DE SMBOL (A X) (COND
    ((OR NODGNV (NOT (EQCAR X 'PLUS))) (COPYL X))
    (T	(PUSHQ TNKEEP!* A)
%PRTLF "%v is a sum. It is replaced by the symbol %v with prt name '%w'"
(PRTLF "%v is a SUM,inside neg power! Thus %w kept as symbol until (ON NODGNV)"
	 A A)
	A) ))

  (DE TEVAL1 (X) (COND
    ((SATOM X) X)				% Atom or SNUMP: (1 . 2)
    ((TENP (CAR X)) (DOCOPY (EVAL X)))
    (T (CONS (CAR X) (MAPCAR (CDR X) 'TEVAL1))) ))

  (DE TEVAL!? (X) (SIMPI (TEVAL1 X)))			% "new"

  (OR (GETD 'COFACTOR)	% The cofactor of element TJK = (TSR INDX1 INDX2).
      (DE COFACTOR (TN I1 I2) (COFACDM TN I1 I2 (GETDIM1 TN))) )


(FLUID '(NR MINUS!* NOSEQUENCE))
(DE EPS11 (X) (AND (GREATERP NR X) (SETQ MINUS!* (NOT MINUS!*))))
(DE EPS1L (NR L) (COND 
   ((NOT L) NIL)
   ((MEMQ NR L))
   ((EPS1L (CAR L) (CDR L)))
   ((MAPC L (FUNCTION EPS11))) ))

(DE EPSMI (A L) (PROG (MINUS!*)
   (RETURN (COND ((EPS1L A L) 0) (MINUS!* -1) (1))) ))

%(DE EPS4 (A B C D) (PROG (MINUS!*)
%   (RETURN (COND ((EPS1L A (LIST B C D)) 0) (MINUS!* -1) (1)))))

(DM EPS4 (L) (LIST 'EPSMI (CADR L)	% Macro,for all EPSL to share.
		%   (MAPCAR (CDDR L) 'EVAL)
		    (CONS 'LIST (CDDR L)) ))

(DECLT (EPS4 (A 1 TO 4)))
% Add all to KONSTANTENSORS would make it too long. User should add,if needed.
(MAPC (SETQ EPSL '(EPS2 EPS3 EPS4 EPS5 EPS6 EPS7 EPS8 EPS9))
      (LAMBDAX (PROGN
	(PUT X 'NAM (SHPMAC '((!e)) '(e)))
	(PUT X 'PRT 'PRTEN)
%	(PUT X 'NCOPY T)
	(PUT X 'PRQFN '(EPSMI EPS1L EPS11))
	(PUT X 'SYMOP  (GET 'EPS4 'SYMOP))
	(MOVD X 'EPS4)	)) )

(DECLT (RDEL S12) (EPS2 A12) (EPS3 (A 1 TO 3)))	% Overwrite SYMOPS on EPS2,3
(OR (GETD 'DEL) (DE DEL (A B) (COND ((EQ A B) 1) (T 0))) )
(DE RDEL (A B) (IFPE (EQ A B) 0 1))	%:h Reversed DEL, 1-DEL

%:hlp Vector PMMM is (+1,-1,-1,...) in any dimension:
(DE PMMM (X) (IFPE (ZEROP X) 1 -1))

%: Thus Minkovski-metric (eta) can be specified: (PDEF GDD)<DEL A B><PMMM @B>$

%:hlp IFNEED - Declare tensor to only have NEEDED components computed
%:h	-by:	(DECLT (TSR (IFNEED T)))   or   (PDEF TSR (IFNEED T)) ...

(DE IFNEED (NOSEQUENCE IL)	% Hlp-fn used in BLFs,to set the switch locally
  (DOCOPY (GETEN NOSEQUENCE IL)) )  % for this tsr,so GETEN makes tsr-component

 ))		% ^^^^ END OF COND ^^^^

% QTCOMF.SL end
